@@ -22,7 +22,7 @@
 /* 3rd party vendors */
 #include <netcdf.h> /* netCDF definitions and C library */
 #ifdef NC_HAVE_META_H
-# include <netcdf_meta.h> /* NC_VERSION_..., NC_HAVE_RENAME_GRP */
+# include <netcdf_meta.h> /* NC_VERSION_..., NC_HAS_QUANTIZE, NC_HAS_ZSTD */
 #endif /* !NC_HAVE_META_H */
 
 /* Personal headers */
@@ -46,6 +46,13 @@
 # define X_CST_PTR_CST_PTR_Y(x,y) x * const * y
 #endif /* !__cplusplus */
 
+#define NCO_ERR_FL_LN do {						\
+    (void)fflush(stdout); /* Synchronize stdout with stderr */		\
+    (void)fprintf(stderr, "Failure in NCO function in file %s at line %d\n",__FILE__,__LINE__); \
+    (void)fflush(stderr);						\
+    return 2;								\
+  } while (0)
+
 #ifdef __cplusplus
 /* Use C-bindings so C++-compiled and C-compiled libraries are compatible */
 extern "C" {
@@ -78,7 +85,7 @@ extern "C" {
 # ifndef bool
 #  define bool int
 # endif /* bool */
-#endif /* __cplusplus */
+#endif /* !__cplusplus */
 #ifndef True
 # define True 1
 #endif /* True */
@@ -108,6 +115,15 @@ extern "C" {
 # define M_PI_2      1.57079632679489661923132169163975144   /* pi/2           */
 #endif /* M_PI_2 */
 
+/* Declare NCO versions (for namespace safety) of MIN/MAX macros in sys/param.h */
+#ifndef NCO_MIN
+# define NCO_MIN(a, b) ((a) < (b) ? (a) : (b))
+#endif /* !NCO_MIN */
+
+#ifndef NCO_MAX
+# define NCO_MAX(a, b) ((a) > (b) ? (a) : (b))
+#endif /* !NCO_MAX */
+
   /* Variables marked CEWI "Compiler Error Warning Initialization" are initialized
      to prevent spurious "warning: `float foo' might be used uninitialized in 
      this function" warnings when, e.g., GCC -Wuninitialized is turned on.
@@ -235,12 +251,12 @@ extern "C" {
 #endif /* NCO_MSS_VAL_SNG */
 
   /* Prototype global functions before defining them in next block */
+  char *nco_cmp_glb_get(void); /* [sng] Global compression specification */
   char *nco_mss_val_sng_get(void); /* [sng] Missing value attribute name */
   char *nco_mta_dlm_get(void); /* [sng] Multi-argument delimiter */
   char *nco_not_mss_val_sng_get(void); /* [sng] Not missing value attribute name */
   char *nco_prg_nm_get(void);
   int nco_prg_id_get(void);
-  int nco_flt_glb_get(void);
   unsigned short nco_baa_cnv_get(void);
   unsigned short nco_bnr_cnv_get(void);
   unsigned short nco_dbg_lvl_get(void);
@@ -248,6 +264,7 @@ extern "C" {
   unsigned short nco_mrd_cnv_get(void);
   unsigned short nco_rth_cnv_get(void);
   unsigned short nco_upk_cnv_get(void);
+  void nco_cmp_glb_set(char *nco_cmp_glb_arg); /* [sng] Global compression specification */
   void nco_fmt_xtn_set(unsigned short nco_fmt_xtn_arg);
   void nco_mta_dlm_set(char *nco_mta_dlm_arg);
 
@@ -266,8 +283,7 @@ extern "C" {
     return nco_mta_dlm;} /* [sng] Multi-argument delimiter */
   void nco_mta_dlm_set(char *nco_mta_dlm_arg){nco_mta_dlm=nco_mta_dlm_arg;} /* [sng] Multi-argument delimiter */
   
-  //unsigned short nco_baa_cnv=0; /* [enm] Bit-Adjustment Algorithm defaults to Bit Grooming (until 20200715) */
-  unsigned short nco_baa_cnv=5; /* [enm] Bit-Adjustment Algorithm defaults to Bit Rounding (after 20200715) */
+  unsigned short nco_baa_cnv=4; /* [enm] Bit-Adjustment Algorithm defaults to Granular BitRound (after 20211003) */
   unsigned short nco_baa_cnv_get(void){return nco_baa_cnv;} /* [enm] Bit-Adjustment Algorithm */
 
   unsigned short nco_bnr_cnv=0; /* [enm] Binary byte-ordering convention to produce (native or byte-swapped) */
@@ -276,8 +292,9 @@ extern "C" {
   unsigned short nco_dbg_lvl=0; /* [enm] Debugging level */
   unsigned short nco_dbg_lvl_get(void){return nco_dbg_lvl;} /* [enm] Debugging level */
 
-  int nco_flt_glb=0; /* [enm] Filter enum */
-  int nco_flt_glb_get(void){return nco_flt_glb;} /* [enm] Filter enum */
+  char *nco_cmp_sng_glb=NULL; /* [sng] Global compression specification */
+  char *nco_cmp_glb_get(void){return nco_cmp_sng_glb;} /* [sng] Global compression specification */
+  void nco_cmp_glb_set(char *nco_cmp_glb_arg){nco_cmp_sng_glb=nco_cmp_glb_arg;} /* [sng] Global compression specification */
 
   unsigned short nco_fmt_xtn=0; /* [enm] Extended file format */
   unsigned short nco_fmt_xtn_get(void){return nco_fmt_xtn;} /* [enm] Extended file format */
@@ -350,23 +367,23 @@ extern "C" {
   /* NCO meta-information available in NCO_VERSION token since forever
      20141008: Define NCO tokens consistent with new netcdf_meta.h tokens */
 #ifndef NCO_VERSION_MAJOR
-# define NCO_VERSION_MAJOR 4
+# define NCO_VERSION_MAJOR 5
 #endif /* !NCO_VERSION_MAJOR */
 #ifndef NCO_VERSION_MINOR
-# define NCO_VERSION_MINOR 9
+# define NCO_VERSION_MINOR 1
 #endif /* !NCO_VERSION_MINOR */
 #ifndef NCO_VERSION_PATCH
 # define NCO_VERSION_PATCH 9
 #endif /* !NCO_VERSION_PATCH */
 #ifndef NCO_VERSION_NOTE
-# define NCO_VERSION_NOTE ".beta02" /* Blank for final versions, non-blank (e.g., "-beta37") for pre-release versions */
+# define NCO_VERSION_NOTE "" /* Blank for final versions, non-blank (e.g., "-beta37") for pre-release versions */
 #endif /* !NCO_VERSION_NOTE */
 #ifndef NCO_LIB_VERSION
   /* Define NC_LIB_VERSION as three-digit number for arithmetic comparisons by CPP */
 # define NCO_LIB_VERSION ( NCO_VERSION_MAJOR * 100 + NCO_VERSION_MINOR * 10 + NCO_VERSION_PATCH )
 #endif /* !NCO_LIB_VERSION */
 #ifndef NCO_VERSION
-# define NCO_VERSION "4.9.9"
+# define NCO_VERSION "5.1.9"
 #endif /* !NCO_VERSION */
 
 /* Compatibility tokens new to netCDF4 netcdf.h: */
@@ -406,6 +423,18 @@ extern "C" {
 #endif
 #ifndef NC_CONTIGUOUS
 # define NC_CONTIGUOUS (1)
+#endif
+  /* Added ~2019 */
+#ifndef NC_COMPACT
+# define NC_COMPACT (2)
+#endif
+  /* Added 202206 netCDF 4.9.0 */
+#ifndef NC_UNKNOWN_STORAGE
+# define NC_UNKNOWN_STORAGE (3)
+#endif
+  /* Added 202206 netCDF 4.9.0 */
+#ifndef NC_VIRTUAL
+# define NC_VIRTUAL (4)
 #endif
 #ifndef NC_NOCHECKSUM
 # define NC_NOCHECKSUM 0
@@ -509,6 +538,9 @@ extern "C" {
 #endif
 #ifndef NC_FORMATX_DAP4
 # define NC_FORMATX_DAP4    (6)
+#endif
+#ifndef NC_FORMATX_NCZARR
+# define NC_FORMATX_NCZARR  (10)
 #endif
 
   /* Three compatibility tokens from pnetcdf.h introduced to NCO 20140604 
@@ -539,6 +571,10 @@ extern "C" {
 #endif
 #ifndef NC_EDAPSVC
 # define NC_EDAPSVC (-70)      /**< DAP server error */
+#endif
+  /* Backwards-compatibility error codes implemented in 202207 to diagnose problems with NCZarr */
+#ifndef NC_ENCZARR
+# define NC_ENCZARR (-137)    /**< Error at NCZarr layer. */
 #endif
 
   /* Backwards-compatibility error codes implemented in 202004 to workaround netCDF 4.7.4 backwards incompatibility with nc_inq_var_filter()
@@ -715,6 +751,8 @@ extern "C" {
     nco_fmt_xtn_pnetcdf=NC_FORMAT_PNETCDF, /* 4 PnetCDF */
     nco_fmt_xtn_dap2=NC_FORMAT_DAP2, /* 5 DAP2 */
     nco_fmt_xtn_dap4=NC_FORMAT_DAP4, /* 6 DAP4 */
+    /* 20220713: Tokens defined as of netCDF 4.8.0 */
+    nco_fmt_xtn_nczarr=NC_FORMATX_NCZARR, /* 10 NCZARR */
   }; /* end nco_fmt_xtn */
 
   enum nco_mrd_cnv{ /* [enm] Multiple Record Dimension convention: for ncecat and ncpdq */
@@ -730,14 +768,15 @@ extern "C" {
   }; /* end nco_rth_cnv */
 
   enum nco_baa_cnv{ /* [enm] Bit-Adjustment Algorithm to use */
-    nco_baa_grm, /* 0 Bit Groom (NCO default since inception) */
-    nco_baa_shv, /* 1 Bit Shave (option since 20160117) */
-    nco_baa_set, /* 2 Bit Set (option since 20160117) */
-    nco_baa_dgr, /* 3 Digit Rounding (GCD19 option since 20190121) */
-    nco_baa_bg2, /* 4 Bit Groom version 2 dynamic masks (option since 20190203) */
-    nco_baa_rnd, /* 5 Bit rounding */
-    nco_baa_sh2, /* 6 Bit Half-shave  */
-    nco_baa_gbg, /* 7 Brute force granular bit grooming */
+    nco_baa_btg, /* 0 BitGroom (option since 201502XX, default 201502XX--20200717) */
+    nco_baa_shv, /* 1 BitShave (option since 20160117) */
+    nco_baa_set, /* 2 BitSet (option since 20160117) */
+    nco_baa_dgr, /* 3 DigitRound (DCG19 option since 20210928, default 20210928--20211003) */
+    nco_baa_gbr, /* 4 Granular BitRound (option since 20211003, default 20211003--present) */
+    nco_baa_bgr, /* 5 BitGroomRound (option since 20200717, default 20200717--20210928) */
+    nco_baa_sh2, /* 6 BitHalf-shave (option since 20200715) */
+    nco_baa_brt, /* 7 BruteForce (option since 20210428) */
+    nco_baa_btr, /* 8 BitRound (option since 20211215 */
   }; /* end nco_baa_cnv */
 
   enum nco_bnr_cnv{ /* [enm] Binary byte-ordering convention to employ (native or byte-swapped) */
@@ -759,7 +798,8 @@ extern "C" {
     aed_delete,
     aed_modify,
     aed_nappend,
-    aed_overwrite
+    aed_overwrite,
+    aed_prepend
   } aed_enm; /* end aed enum */
   
   typedef enum gpe{ /* [enm] Group Path Editing mode */
@@ -1097,7 +1137,7 @@ extern "C" {
   typedef enum nco_wgt_typ_enm{
     nco_wgt_nil=0,
     nco_wgt_con, /* Conservative algorithm */
-    nco_wgt_dwe, /* Distance-weighted extrapolation */
+    nco_wgt_idw, /* Distance-weighted extrapolation */
     nco_wgt_bln  /* Bilinear interpolation */
   } nco_wgt_typ_enm;
 
@@ -1141,7 +1181,8 @@ typedef enum nco_xtr_typ_enm
     char *fl_out; /* [sng] File containing regridded fields */
     char *fl_out_tmp; /* [sng] Temporary file containing regridded fields */
     char *fl_map; /* [sng] File containing mapping weights from source to destination grid */
-    char *fl_vrt; /* [sng] File containing vertical coordinate grid */
+    char *fl_vrt_in; /* [sng] File containing input vertical coordinate grid */
+    char *fl_vrt_out; /* [sng] File containing output vertical coordinate grid */
     char *fl_msh;  /* [sng] File that contains SCRIP formatted overlap mesh */
     char *var_nm; /* [sng] Variable for special regridding treatment */
     // Grid names specifiable with key-value syntax
@@ -1151,11 +1192,15 @@ typedef enum nco_xtr_typ_enm
     char *bnd_tm_nm; /* [sng] Name of dimension to employ for temporal bounds */
     char *col_nm_in; /* [sng] Name to recognize as input horizontal spatial dimension on unstructured grid */
     char *col_nm_out; /* [sng] Name of horizontal spatial output dimension on unstructured grid */
+    char *dpt_nm_in; /* [sng] Name of input variable to recognize as depth for depth/height grids */
+    char *dpt_nm_out; /* [sng] Name of variable to output as depth for depth/height grids */
+    char *dpt_nm_tpl; /* [sng] Name of template variable to recognize as depth for depth/height grids */
     char *fl_hnt_dst; /* [sng] ERWG hint destination */
     char *fl_hnt_src; /* [sng] ERWG hint source */
     char *frc_nm; /* [sng] Name of variable containing gridcell fraction */
     char *ilev_nm_in; /* [sng] Name of input dimension to recognize as vertical dimension at layer interfaces */
     char *ilev_nm_out; /* [sng] Name of output vertical dimension at layer interfaces */
+    char *ilev_nm_tpl; /* [sng] Name of template vertical dimension at layer interfaces */
     char *lat_bnd_nm; /* [sng] Name of rectangular boundary variable for latitude */
     char *lat_dmn_nm; /* [sng] Name of latitude dimension in inferred grid */
     char *lat_nm_in; /* [sng] Name of input dimension to recognize as latitude */
@@ -1164,15 +1209,22 @@ typedef enum nco_xtr_typ_enm
     char *lat_wgt_nm; /* [sng] Name of variable containing latitude weights */
     char *lev_nm_in; /* [sng] Name of input dimension to recognize as vertical dimension at layer midpoints */
     char *lev_nm_out; /* [sng] Name of output vertical dimension at layer midpoints */
+    char *lev_nm_tpl; /* [sng] Name of template vertical dimension at layer midpoints */
     char *lon_bnd_nm; /* [sng] Name of rectangular boundary variable for longitude */
     char *lon_dmn_nm; /* [sng] Name of longitude dimension in inferred grid */
     char *lon_nm_in; /* [sng] Name of dimension to recognize as longitude */
     char *lon_nm_out; /* [sng] Name of output dimension for longitude */
     char *lon_vrt_nm; /* [sng] Name of non-rectangular boundary variable for longitude */
     char *msk_nm; /* [sng] Name of variable containing destination mask */
-    char *plev_nm_in; /* [sng] Name of variable to recognize as vertical coordinate for pure pressure grids */
+    char *plev_nm_in; /* [sng] Name of input variable to recognize as vertical coordinate for pure pressure grids */
+    char *plev_nm_out; /* [sng] Name of variable to output as vertical coordinate for pure pressure grids */
+    char *plev_nm_tpl; /* [sng] Name of template variable to recognize as vertical coordinate for pure pressure grids */
+    char *ps_nm_in; /* [sng] Name of input variable to recognize as surface pressure for hybrid/sigma pressure grids */
+    char *ps_nm_out; /* [sng] Name of variable to output as surface pressure for hybrid/sigma pressure grids */
+    char *ps_nm_tpl; /* [sng] Name of template variable to recognize as surface pressure for hybrid/sigma pressure grids */
     char *sgs_frc_nm; /* [sng] Name of variable containing sub-gridscale fraction */
     char *sgs_msk_nm; /* [sng] Name of variable containing sub-gridscale mask */
+    char *vrt_in; /* [sng] Name of file containing vertical grid coefficients for input file */
     char *vrt_nm; /* [sng] Name of dimension to employ for vertices */
     // User-specified grid properties
     char *fl_grd; /* [sng] Name of SCRIP grid file to create */
@@ -1214,6 +1266,7 @@ typedef enum nco_xtr_typ_enm
     int area_mth; /* [enm]  area method used calculating grid cell area */
     long idx_dbg; /* [idx] Index of gridcell for debugging */
     long tst; /* [enm] Generic key for testing (undocumented) */
+    nco_bool flg_add_fll; /* [flg] Add _FillValue to fields with empty destination cells */
     nco_bool flg_area_out; /* [flg] Add area to output */
     nco_bool flg_cll_msr; /* [flg] Add cell_measures attribute */
     nco_bool flg_crv; /* [flg] Use curvilinear coordinates */
@@ -1224,8 +1277,11 @@ typedef enum nco_xtr_typ_enm
     nco_bool flg_grd; /* [flg] Create SCRIP-format grid file */
     nco_bool flg_grd_dst; /* [flg] User-specified destination grid */
     nco_bool flg_grd_src; /* [flg] User-specified input grid */
+    nco_bool flg_mpt_mss; /* [flg] Set empty (sgs_frc=0.0) SGS cells to missing value */
+    nco_bool flg_msk_apl; /* [flg] Apply msk_out to variables after regridding */
     nco_bool flg_msk_out; /* [flg] Add mask to output */
     nco_bool flg_nfr; /* [flg] Infer SCRIP-format grid file */
+    nco_bool flg_ps_rtn; /* [flg] Retain surface pressure variable in vertical interpolation output */
     nco_bool flg_rnr; /* [flg] Renormalize destination values by valid area */
     nco_bool flg_s1d; /* [flg] Unpack sparse-1D CLM/ELM variables */
     nco_bool flg_stg; /* [flg] Write staggered grid with FV output */
@@ -1651,7 +1707,7 @@ typedef enum nco_xtr_typ_enm
 
 
 #ifdef __cplusplus
-} /* end extern "C" */
+} /* !extern "C" */
 #endif /* !__cplusplus */
 
-#endif /* NCO_H */
+#endif /* !NCO_H */
